

set -ex



scrapy version -v
pip install setuptools-scm
pip install mitmproxy pytest-twisted sybil
mv scrapy _scrapy
rm tests/test_command_parse.py tests/test_commands.py
rm -f tests/test_command_parse.py tests/test_commands.py tests/test_crawler.py tests/test_downloader_handlers.py
pytest _scrapy tests
exit 0
